/*
 * Compresses a fastq stream to fqz format.
 * It only understands the simplified 1 line per sequence format.
 *
 * Usage: fastq2fqz < a.fastq > a.fqz
 */

#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <zlib.h>

//#include <io_lib/deflate_interlaced.h>
#include "deflate_interlaced.h"

#define VIA_IOLIB 666


// Somewhat pointless, but use IOLIB for huffman only and then gzip
// to try and spot repeated strings in the huffman output (hard, but possible)
// works ok as a faster alternative to pure gzip.
//
//#define GZIP_OUT


// We can eiher use io_lib deflate_interlaced.c for huffman encoding, or
// zlib for a full LZ77 implementation. The io_lib version also allows for
// interlaced deflate which improves compression ratio for quality values.
// In theory (and sometimes in practice?) we can use io_lib's implementation
// for pure huffman encoding and use zlib as the decoder in fqz2fastq.
// The io_lib huffman encoder is approx 4x faster than zlib's, but
// conversely zlib's decoder is faster (although not by the same
// degree). NB: if you use io_lib for encoding and zlib for decoding
// you'll need to change the window_size from 15 to -15 in the deflateInit2
// and inflateInit2 calls.
//
// LZ77 level 5 and 7 works ok too. Beyond that it's minimal returns
// for high CPU

//#define NAME_STRAT VIA_IOLIB
#define NAME_STRAT Z_FILTERED
//#define NAME_STRAT Z_HUFFMAN_ONLY
#define NAME_LEVEL 3

//#define SEQ_STRAT VIA_IOLIB
#define SEQ_STRAT Z_FILTERED
//#define SEQ_STRAT Z_HUFFMAN_ONLY
#define SEQ_LEVEL 3

#define QUAL_STRAT VIA_IOLIB
//#define QUAL_STRAT Z_RLE
#define QUAL_LEVEL 1

static void zlib_mem_deflate(block_t *blk, char *data, size_t size,
			     int level, int strat) {
    z_stream s;
    unsigned char *cdata = NULL; /* Compressed output */
    int cdata_alloc = 0;
    int cdata_pos = 0;
    int err;

    cdata = malloc(cdata_alloc = size*1.05+10);
    cdata_pos = 0;

    /* Initialise zlib stream */
    s.zalloc = Z_NULL; /* use default allocation functions */
    s.zfree  = Z_NULL;
    s.opaque = Z_NULL;
    s.next_in  = (unsigned char *)data;
    s.avail_in = size;
    s.total_in = 0;
    s.next_out  = cdata;
    s.avail_out = cdata_alloc;
    s.total_out = 0;
    s.data_type = Z_BINARY;

    err = deflateInit2(&s, level, Z_DEFLATED, 15, 9, strat);

    /* Encode to 'cdata' array */
    for (;s.avail_in;) {
	s.next_out = &cdata[cdata_pos];
	s.avail_out = cdata_alloc - cdata_pos;
	if (cdata_alloc - cdata_pos <= 0) {
	    fprintf(stderr, "Deflate produced larger output than expected. Abort\n"); 
	    return;
	}
	err = deflate(&s, Z_NO_FLUSH);
	cdata_pos = cdata_alloc - s.avail_out;
	if (err != Z_OK) {
	    fprintf(stderr, "zlib deflate error: %s\n", s.msg);
	    break;
	}
    }
    if (deflate(&s, Z_FINISH) != Z_STREAM_END) {
	fprintf(stderr, "zlib deflate error: %s\n", s.msg);
    }
    //*cdata_size = s.total_out;

    if (deflateEnd(&s) != Z_OK) {
	fprintf(stderr, "zlib deflate error: %s\n", s.msg);
    }

    //return (char *)cdata;

    store_bytes(blk, cdata, s.total_out);
    free(cdata);
}

static void zlib_mem_deflate2(block_t *blk, char *data, size_t size,
			     int level, int strat, int chunks) {
    z_stream s;
    unsigned char *cdata = NULL; /* Compressed output */
    int cdata_alloc = 0;
    int cdata_pos = 0;
    int err;

    cdata = malloc(cdata_alloc = size*1.05+10);
    cdata_pos = 0;

    /* Initialise zlib stream */
    s.zalloc = Z_NULL; /* use default allocation functions */
    s.zfree  = Z_NULL;
    s.opaque = Z_NULL;
    s.next_in  = (unsigned char *)data;
    s.avail_in = size <= chunks ? size : chunks;
    s.total_in = 0;
    s.next_out  = cdata;
    s.avail_out = cdata_alloc;
    s.total_out = 0;
    s.data_type = Z_BINARY;

    err = deflateInit2(&s, level, Z_DEFLATED, 15, 9, strat);

    /* Encode to 'cdata' array */
    do {
	s.avail_in = size <= chunks ? size : chunks;
	size -= s.avail_in;

	for (;s.avail_in;) {
	    s.next_out = &cdata[cdata_pos];
	    s.avail_out = cdata_alloc - cdata_pos;
	    if (cdata_alloc - cdata_pos <= 0) {
		fprintf(stderr, "Deflate produced larger output than expected. Abort\n"); 
		return;
	    }
	    err = deflate(&s, Z_SYNC_FLUSH);
	    cdata_pos = cdata_alloc - s.avail_out;
	    if (err != Z_OK) {
		fprintf(stderr, "zlib deflate error: %s\n", s.msg);
		break;
	    }
	}
    } while (size);
    if (deflate(&s, Z_FINISH) != Z_STREAM_END) {
	fprintf(stderr, "zlib deflate error: %s\n", s.msg);
    }
    //*cdata_size = s.total_out;

    if (deflateEnd(&s) != Z_OK) {
	fprintf(stderr, "zlib deflate error: %s\n", s.msg);
    }

    //return (char *)cdata;

    store_bytes(blk, cdata, s.total_out);
    free(cdata);
}

/*
 * Compared against 1Mb of block size:
 *     5Mb   => ~1% smaller
 *   100Kb   => 7-8% larger
 */
#define BLK_SIZE 2000000
//#define BLK_SIZE 100000
#define MAX_LINE_LEN 1024

char *encode_data(char *in, int in_len, int *out_len, char **in_end) {
    char last_name[MAX_LINE_LEN];
    char name[MAX_LINE_LEN];
    char seq[MAX_LINE_LEN];
    char qual[MAX_LINE_LEN];

    char *seq_buf = malloc(in_len);
    char *qual_buf = malloc(in_len);
    char *name_buf = malloc(in_len);
    char *nptr, *qptr, *sptr;
    char nseq_buf[4];
    unsigned char L[256];
    char *ret;

    int slen, nlen, last_nlen = 0;
    int s_ind, q_ind, n_ind, s_ind3 = 0;
    int i, j, end = 0;

#if NAME_STRAT == VIA_IOLIB
    huffman_codeset_t *cds_name;
#endif

#if SEQ_STRAT == VIA_IOLIB
    huffman_codeset_t *cds_seq;
#endif

#if QUAL_STRAT == VIA_IOLIB
    huffman_codeset_t *cds_qual;
#endif
    block_t *blk;

    int nseqs = 0;

    memset(L, 4, 256);
    L['A'] = L['a'] = 0;
    L['C'] = L['c'] = 1;
    L['G'] = L['g'] = 2;
    L['T'] = L['t'] = 3;

    /* ---- part 1: Load whole sequences ---- */
    for (n_ind = s_ind = q_ind = i = 0; i < in_len; ) {
	/* ---- Fetch a whole sequence record. */
	
	/* Name */
	if (in[i] != '@')
	    return NULL;

	for (j = 0, i++; i < in_len && in[i] != '\n'; i++, j++)
	    name[j] = in[i];
	if (++i >= in_len)
	    break;
	nlen = j;

	/* sequence */
	for (j = 0; i < in_len && in[i] != '\n'; i++, j++)
	    seq[j] = in[i];
	if (++i >= in_len)
	    break;
	slen = j;

	/* name2 */
	if (in[i] != '+')
	    return NULL;

	for (j = 0; i < in_len && in[i] != '\n'; i++, j++)
	    ;
	if (++i >= in_len)
	    break;

	/* quality */
	for (j = 0; i < in_len && in[i] != '\n'; i++, j++)
	    qual[j] = in[i];
	if (++i > in_len)
	    break;
	if (j != slen)
	    return NULL;

	nseqs++;
	end = i;
	

	/* ---- Append to name, sequence and quality buffers */
	for (j = 0; j < nlen && j < last_nlen; j++) {
	    if (name[j] != last_name[j])
		break;
	}
	if (!isprint(j) && n_ind && name_buf[n_ind-1] == 0)
	    n_ind--;
	name_buf[n_ind++] = j;
	for (; j < nlen; j++) {
	    name_buf[n_ind++] = name[j];
	}
	name_buf[n_ind++] = 0; /* Delineates names */
	memcpy(last_name, name, nlen);
	last_nlen = nlen;

#if 1
	for (j = 0; j < slen; j+=3) {
	    seq_buf[s_ind++] = L[(unsigned char)seq[j]]*25
		+ (j+1<slen ? L[(unsigned char)seq[j+1]]*5 : 0)
		+ (j+2<slen ? L[(unsigned char)seq[j+2]] : 0);
	}

	/*/

	for (j = 0; j < slen; j++) {
	    seq_buf[s_ind] = seq_buf[s_ind]*5 + L[(unsigned char)seq[j]];
	    if (++s_ind3 == 3) {
		s_ind3 = 0;
		s_ind++;
	    }
	}
	/**/
#else
	for (j = 0; j < slen; j++) {
	    seq_buf[s_ind++] = seq[j];
	}
#endif

	for (j = 0; j < slen; j++) {
	    qual_buf[q_ind++] = qual[j];
	}
    }

    *in_end = in + end;

    /*
     * Optionally rotate the quality buffer.
     * This coupled with a zlib chunked deflate to split on boundaries
     * is directly equivalent to the io_lib implementation of
     * interlaced-deflate. It's pehaps not as CPU efficient though.
     */
    if (0) {
	int i, x = 0;
	char *rq = malloc(q_ind);

	for (j = 0; j < slen; j++) {
	    for (i = 0; i < nseqs; i++) {
		rq[x++] = qual_buf[i*slen+j];
	    }
	}
	free(qual_buf);
	qual_buf = rq;
    }

    /* ---- Compute huffman trees for each component ---- */
#if NAME_STRAT == VIA_IOLIB
    cds_name = generate_code_set(CODE_USER+0, 1, name_buf, n_ind,
				 nseqs, MAX_CODE_LEN, 0);
#endif

#if SEQ_STRAT == VIA_IOLIB
    cds_seq  = generate_code_set(CODE_USER+1, 1, seq_buf,  s_ind,
				 nseqs, MAX_CODE_LEN, 0);
#endif

#if QUAL_STRAT == VIA_IOLIB
    /*
     * This code assumes a constant sequence length, slen.
     * We produce one huffman tree per base in the sequence, acknowledging
     * that the distributions of qualities at the start look different
     * than at the end. This can shrink the overall file size by 5-10%.
     *
     * Specify "1" instead of "slen" if this assumption is false.
     */
    cds_qual = generate_code_set(CODE_USER+2, slen, qual_buf, q_ind,
    				 nseqs, MAX_CODE_LEN, 0);
#endif

    /* ---- Compress it! ---- */

    blk = block_create(NULL, 8192);

    nseq_buf[0] = (nseqs >>  0) & 0xff;
    nseq_buf[1] = (nseqs >>  8) & 0xff;
    nseq_buf[2] = (nseqs >> 16) & 0xff;
    nseq_buf[3] = (nseqs >> 24) & 0xff;
    store_bytes(blk, nseq_buf, 4);

    nseq_buf[0] = (slen >>  0) & 0xff;
    nseq_buf[1] = (slen >>  8) & 0xff;
    nseq_buf[2] = (slen >> 16) & 0xff;
    nseq_buf[3] = (slen >> 24) & 0xff;
    store_bytes(blk, nseq_buf, 4);

    nptr = name_buf;
    sptr = seq_buf;
    qptr = qual_buf;

#if NAME_STRAT == VIA_IOLIB
    store_codes(blk, cds_name, 1);
    huffman_multi_encode(blk, cds_name, CODE_USER+0, name_buf, n_ind);
    huffman_codeset_destroy(cds_name);
    /*
     * zlib always expects inflate to start on a byte boundary, so if
     * we want to mix and match between encoder and decoder for optimal
     * speed we need to force the same with io_lib's implementation too.
     */
    if (blk->bit) {
	blk->byte++;
	blk->bit = 0;
    }
#else
    zlib_mem_deflate(blk, name_buf, n_ind, NAME_LEVEL, NAME_STRAT);
#endif

#if SEQ_STRAT == VIA_IOLIB
    store_codes(blk, cds_seq,  1);
    huffman_multi_encode(blk, cds_seq,  CODE_USER+1, seq_buf,  s_ind);
    huffman_codeset_destroy(cds_seq);
    if (blk->bit) {
	blk->byte++;
	blk->bit = 0;
    }
#else
    zlib_mem_deflate(blk, seq_buf,  s_ind, SEQ_LEVEL, SEQ_STRAT);
#endif

#if QUAL_STRAT == VIA_IOLIB
    store_codes(blk, cds_qual, 1);
    huffman_multi_encode(blk, cds_qual, CODE_USER+2, qual_buf, q_ind);
    huffman_codeset_destroy(cds_qual);
    if (blk->bit) {
	blk->byte++;
	blk->bit = 0;
    }
#else
    zlib_mem_deflate(blk, qual_buf, q_ind, QUAL_LEVEL, QUAL_STRAT);
    //zlib_mem_deflate2(blk, qual_buf, q_ind, QUAL_LEVEL, QUAL_STRAT, nseqs);
#endif

    *out_len = blk->byte + (blk->bit != 0);

    ret = blk->data;
    block_destroy(blk,  1);

    if (seq_buf)  free(seq_buf);
    if (qual_buf) free(qual_buf);
    if (name_buf) free(name_buf);

    return ret;
}

char *mem_gz(char *in, int in_len, int *out_len) {
    int err;
    z_stream s;
    char *out;
    int gz_alloc = 0, gz_pos = 0;

    s.zalloc = (alloc_func)0;
    s.zfree = (free_func)0;
    s.opaque = (voidpf)0;
    s.next_in = (unsigned char *)in;
    s.avail_in = in_len;

    err = deflateInit2(&s, 6, Z_DEFLATED, -MAX_WBITS,
		       9 /* DEF_MEM_LEVEL */, Z_FILTERED);
    if (err != Z_OK)
	exit(1);
    out = (char *)malloc(gz_alloc = in_len*1.02+20);
    for (;s.avail_in;) {
	s.next_out = (unsigned char *)&out[gz_pos];
	s.avail_out = gz_alloc - gz_pos;
	if (gz_alloc - gz_pos <= 0) {
	    exit(1);
	}
	err = deflate(&s, Z_NO_FLUSH);
	gz_pos = gz_alloc - s.avail_out;
	if (err != Z_OK)
	    exit(1);
    }
    deflate(&s, Z_FINISH);
    deflateEnd(&s);

    *out_len = s.total_out;
    return out;
}

static char data[BLK_SIZE];
int main(void) {
    int sz, blk_start = 0;

    /* Get a block at a time, trimmed back to a whole FASTQ entry */
    blk_start = 0;
    while ((sz = read(0, &data[blk_start], BLK_SIZE - blk_start)) > 0) {
	char *comp, *data_new, *gz_comp;
	int comp_len;
	unsigned char len_buf[4];

	comp = encode_data(data, sz + blk_start, &comp_len, &data_new);
	if (comp == NULL) {
	    fprintf(stderr, "Abort: encode_data returned NULL\n");
	    exit(1);
	}

	/* Now gzip it too */
#ifdef GZIP_OUT
	gz_comp = mem_gz(comp, comp_len, &comp_len);
	free(comp);
	comp = gz_comp;
#endif

	len_buf[0] = (comp_len >>  0) & 0xff;
	len_buf[1] = (comp_len >>  8) & 0xff;
	len_buf[2] = (comp_len >> 16) & 0xff;
	len_buf[3] = (comp_len >> 24) & 0xff;
	write(1, len_buf, 4);
	write(1, comp, comp_len);

	free(comp);

	/* We maybe ended on a partial fastq entry, so start from there */
	memmove(data, data_new, (sz + blk_start) - (data_new - data));
	blk_start = (sz + blk_start) - (data_new - data);
    }

    return 0;
}
