/*
 * Uncompresses an fqz file to produce fastq format
 */

#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <zlib.h>

//#include <io_lib/deflate_interlaced.h>
#include "deflate_interlaced.h"

// See fastq2fqz.c for encoding params. The file format doesn't list
// these so the two need to be kept in sync.

//#define GZIP_OUT

//#define NAME_VIA_IOLIB
//#define SEQ_VIA_IOLIB
#define QUAL_VIA_IOLIB

/*
 * Only operates on byte boundaries of blk_in. (ie blk_in->bit == 0)
 */
static block_t *zlib_mem_inflate(block_t *blk_in) {
    char *cdata = &blk_in->data[blk_in->byte];
    size_t csize = blk_in->alloc - blk_in->byte;
    block_t *blk_out;
    z_stream s;
    unsigned char *data = NULL; /* Uncompressed output */
    int data_alloc = 0;
    int err;

    /* Starting point at uncompressed size, 4x compressed */
    data = malloc(data_alloc = csize*4+10);

    /* Initialise zlib stream */
    s.zalloc = Z_NULL; /* use default allocation functions */
    s.zfree  = Z_NULL;
    s.opaque = Z_NULL;
    s.next_in  = (unsigned char *)cdata;
    s.avail_in = csize;
    s.total_in = 0;
    s.next_out  = data;
    s.avail_out = data_alloc;
    s.total_out = 0;

    err = inflateInit(&s);
    //err = inflateInit2(&s, -15);

    /* Decode to 'data' array */
    for (;s.avail_in;) {
	s.next_out = &data[s.total_out];
	err = inflate(&s, Z_NO_FLUSH);
	if (err == Z_STREAM_END)
	    break;

	if (err != Z_OK) {
	    fprintf(stderr, "zlib inflate error: %s\n", s.msg);
	    break;
	}

	/* More to come, so realloc */
	data = realloc(data, data_alloc += s.avail_in*4 + 10);
	s.avail_out += s.avail_in*4+10;
    }
    inflateEnd(&s);

    blk_in->byte += s.total_in;
    blk_out = block_create(data, s.total_out);
    blk_out->byte = s.total_out;
    return blk_out;
}

/*
 * ---------------------------------------------------------------------------
 * Debug code. This turns the library into a stand-alone program for
 * easy debugging.x
 */
#define SYM_EOF 256
static void output_codes(FILE *fp, huffman_codes_t *cds) {
    int i, j;
    int nbits_in = 0, nbits_out = 0;
    huffman_code_t *codes = cds->codes;
    int ncodes = cds->ncodes;

    fprintf(fp, "static huffman_code_t codes_FIXME[] = {\n");
    for (i = j = 0; i < ncodes; i++) {
	nbits_out += codes[i].nbits * codes[i].freq;
	nbits_in  += 8*codes[i].freq;
	if (j == 0)
	    fprintf(fp, "    ");
	if (codes[i].symbol == SYM_EOF) {
	    fprintf(fp, "{SYM_EOF,%3d}, ", codes[i].nbits);
	    j = 10;
	} else {
	    if (isalnum(codes[i].symbol)) {
		fprintf(fp, "{'%c',%3d}, ", codes[i].symbol, codes[i].nbits);
	    } else {
		fprintf(fp, "{%3d,%3d}, ", codes[i].symbol, codes[i].nbits);
	    }
	}
	j++;
	
	if (j >= 6) {
	    fputc('\n', fp);
	    j = 0;
	}
    }
    if (j)
	fputc('\n', fp);
    fprintf(fp, "};\n");
    fprintf(fp, "/* Expected compression to %f of input */\n",
	    (double)nbits_out/nbits_in);
}

static void output_codeset(FILE *fp, huffman_codeset_t *cds) {
    int i;
    for (i = 0; i < cds->ncodes; i++) {
	output_codes(fp, cds->codes[i]);
    }
    fprintf(fp, "\n");
}

#define MAX_NAME 80

block_t *decode(block_t *blk_in) {
#ifdef NAME_VIA_IOLIB
    huffman_codeset_t *cs_name = NULL;
#endif
#ifdef SEQ_VIA_IOLIB
    huffman_codeset_t *cs_seq  = NULL;
#endif
#ifdef QUAL_VIA_IOLIB
    huffman_codeset_t *cs_qual = NULL;
#endif
    //block_t *blk_out;
    block_t *blk_name;
    block_t *blk_seq;
    block_t *blk_qual;
    int bfinal;
    int i, j, nseqs, slen;
    int nind, sind, qind;
    char *fastq, *cp;
    char *last_name = NULL;


    /* FIXME: implement a proper get_bytes function */
    nseqs = (blk_in->data[blk_in->byte+0] <<  0) |
	    (blk_in->data[blk_in->byte+1] <<  8) |
	    (blk_in->data[blk_in->byte+2] << 16) |
	    (blk_in->data[blk_in->byte+3] << 24);
    blk_in->byte += 4;

    slen = (blk_in->data[blk_in->byte+0] <<  0) |
	   (blk_in->data[blk_in->byte+1] <<  8) |
	   (blk_in->data[blk_in->byte+2] << 16) |
	   (blk_in->data[blk_in->byte+3] << 24);
    blk_in->byte += 4;

#ifdef NAME_VIA_IOLIB
    cs_name = restore_codes(blk_in, &bfinal);
    blk_name = huffman_multi_decode(blk_in, cs_name);
    //output_codeset(stdout, cs_name);
    huffman_codeset_destroy(cs_name);
    if (blk_in->bit) {
	blk_in->byte++;
	blk_in->bit = 0;
    }
#else
    blk_name = zlib_mem_inflate(blk_in);
#endif

#ifdef SEQ_VIA_IOLIB
    cs_seq  = restore_codes(blk_in, &bfinal);
    blk_seq  = huffman_multi_decode(blk_in, cs_seq);
    //output_codeset(stdout, cs_seq);
    huffman_codeset_destroy(cs_seq);
#else
    blk_seq = zlib_mem_inflate(blk_in);
#endif

#ifdef QUAL_VIA_IOLIB
    cs_qual = restore_codes(blk_in, &bfinal);
    blk_qual = huffman_multi_decode(blk_in, cs_qual);
    //output_codeset(stdout, cs_qual);
    huffman_codeset_destroy(cs_qual);
#else
    blk_qual = zlib_mem_inflate(blk_in);
#endif

    nind = sind = qind = 0;

    cp = fastq = (char *)malloc(nseqs * (MAX_NAME + slen*2 + 4));
    for (i = 0; i < nseqs; i++) {
	char *this_name;

	/* Name */
	*cp++ = '@';

	this_name = cp;
	if (last_name) {
	    for (j = 0; j < blk_name->data[nind]; j++)
		*cp++ = last_name[j];
	}
	for (nind++;
	     nind < blk_name->byte && isprint(blk_name->data[nind]);
	     nind++, j++)
	    *cp++ = blk_name->data[nind];
	if (nind < blk_name->byte && blk_name->data[nind] == 0)
	    nind++;
	*cp++ = '\n';
	last_name = this_name;

	/* Sequence */
#if 1
	for (j = 0; j < slen; j+=3) {
	    unsigned char base = blk_seq->data[sind++];
	    *cp++ = "ACGTN"[base/25];
	    if (j+1 < slen) *cp++ = "ACGTN"[(base/5) % 5];
	    if (j+2 < slen) *cp++ = "ACGTN"[base % 5];
	}
#else
	memcpy(cp, &blk_seq->data[sind], slen);
	sind += slen;
	cp += slen;
#endif

	*cp++ = '\n';
	*cp++ = '+';
	*cp++ = '\n';

	/* Quality */
	memcpy(cp, &blk_qual->data[qind], slen);
	qind += slen;
	cp += slen;
	*cp++ = '\n';
    }

    
    write(1, fastq, cp-fastq);
    free(fastq);

    block_destroy(blk_name, 0);
    block_destroy(blk_seq,  0);
    block_destroy(blk_qual, 0);

    return NULL;
}

char *memgunzip(char *data, size_t size, size_t *udata_size) {
    z_stream s;
    char *udata = NULL;
    int udata_alloc = 0;
    int udata_pos = 0;

    /* Initialise zlib stream starting after the header */
    s.zalloc = (alloc_func)0;
    s.zfree = (free_func)0;
    s.opaque = (voidpf)0;
    s.next_in = (unsigned char *)data;
    s.avail_in = size;
    inflateInit2(&s, -MAX_WBITS);

    /* Decode to 'udata' array */
    for (;;) {
        int err;
        if (udata_alloc - udata_pos < 1) {
            udata_alloc = udata_alloc ? udata_alloc * 2 : 256;
            udata = realloc(udata, udata_alloc);
        }
        s.next_out = (unsigned char *)&udata[udata_pos];
        s.avail_out = udata_alloc - udata_pos;
        err = inflate(&s, Z_NO_FLUSH);
        udata_pos = udata_alloc - s.avail_out;
        if (err) {
	    //err = Z_STREAM_END; /* HACK! */
            if (err == Z_STREAM_END) {
                break;
            } else {
                inflateEnd(&s);
                return NULL;
            }
        }
    }

    inflateEnd(&s);
    *udata_size = udata_pos;

    return udata;
}

/*
 * Reads the next block of compressed data from fp.
 * Returns block_t ptr on success (free with block_destroy)
 *         NULL on failure
 */
block_t *next_block(FILE *fp) {
    unsigned char len_buf[4];
    unsigned int len;
    unsigned char *data;

    if (4 != fread(len_buf, 1, 4, fp))
	return NULL;

    len = (len_buf[0] <<  0) |
	  (len_buf[1] <<  8) | 
	  (len_buf[2] << 16) | 
	  (len_buf[3] << 24);

    if (NULL == (data = (unsigned char *)malloc(len)))
	return NULL;
    if (len != fread(data, 1, len, fp)) {
	free(data);
	return NULL;
    }

    return block_create(data, len);
}

int main(void) {
    block_t *blk, *ublk;

    while ((blk = next_block(stdin))) {
#ifdef GZIP_OUT
	size_t ulen;
	char *ungz;

	ungz = memgunzip(blk->data, blk->alloc, &ulen);
	ublk = block_create(ungz, ulen);
	block_destroy(blk, 0);
#else
	ublk = blk;
#endif

	decode(ublk);
	block_destroy(ublk, 0);
    }

    return 0;
}
